@@ -5,7 +5,7 @@
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 from __future__ import annotations
 
-from typing import Awaitable
+from typing import Awaitable, Literal, overload
 import json
 
 from mautrix.api import Method, Path
@@ -16,6 +16,7 @@ from mautrix.types import (
     ContentURI,
     Event,
     EventContent,
+    EventContext,
     EventID,
     EventType,
     FilterID,
@@ -125,12 +126,83 @@ class EventMethods(BaseClientAPI):
         except SerializerError as e:
             raise MatrixResponseError("Invalid event in response") from e
 
+    async def get_event_context(
+        self,
+        room_id: RoomID,
+        event_id: EventID,
+        limit: int | None = 10,
+        filter: RoomEventFilter | None = None,
+    ) -> EventContext:
+        """
+        Get a number of events that happened just before and after the specified event.
+        This allows clients to get the context surrounding an event, as well as get the state at
+        an event and paginate in either direction.
+
+        Args:
+            room_id: The room to get events from.
+            event_id: The event to get context around.
+            limit: The maximum number of events to return. The limit applies to the total number of
+                   events before and after the requested event. A limit of 0 means no other events
+                   are returned, while 2 means one event before and one after are returned.
+            filter: A JSON RoomEventFilter_ to filter returned events with.
+
+        Returns:
+            The event itself, up to ``limit/2`` events before and after the event, the room state
+            at the event, and pagination tokens to scroll up and down.
+
+        .. _RoomEventFilter:
+            https://spec.matrix.org/v1.1/client-server-api/#filtering
+        """
+        query_params = {}
+        if limit is not None:
+            query_params["limit"] = str(limit)
+        if filter is not None:
+            query_params["filter"] = (
+                filter.serialize() if isinstance(filter, Serializable) else filter
+            )
+        resp = await self.api.request(
+            Method.GET,
+            Path.v3.rooms[room_id].context[event_id],
+            query_params=query_params,
+            metrics_method="get_event_context",
+        )
+        return EventContext.deserialize(resp)
+
+    @overload
     async def get_state_event(
         self,
         room_id: RoomID,
         event_type: EventType,
         state_key: str = "",
-    ) -> StateEventContent:
+        *,
+        format: Literal["content"] = "content",
+    ) -> StateEventContent: ...
+    @overload
+    async def get_state_event(
+        self,
+        room_id: RoomID,
+        event_type: EventType,
+        state_key: str = "",
+        *,
+        format: Literal["event"],
+    ) -> StateEvent: ...
+    @overload
+    async def get_state_event(
+        self,
+        room_id: RoomID,
+        event_type: EventType,
+        state_key: str = "",
+        *,
+        format: str = "content",
+    ) -> StateEventContent | StateEvent: ...
+    async def get_state_event(
+        self,
+        room_id: RoomID,
+        event_type: EventType,
+        state_key: str = "",
+        *,
+        format: str = "content",
+    ) -> StateEventContent | StateEvent:
         """
         Looks up the contents of a state event in a room. If the user is joined to the room then the
         state is taken from the current state of the room. If the user has left the room then the
@@ -142,6 +214,9 @@ class EventMethods(BaseClientAPI):
             room_id: The ID of the room to look up the state in.
             event_type: The type of state to look up.
             state_key: The key of the state to look up. Defaults to empty string.
+            format: The format of the state event to return. Defaults to "content", which only returns
+                the content of the state event. If set to "event", the full event is returned.
+                See https://github.com/matrix-org/matrix-spec/issues/1047 for more info.
 
         Returns:
             The state event.
@@ -149,11 +224,17 @@ class EventMethods(BaseClientAPI):
         content = await self.api.request(
             Method.GET,
             Path.v3.rooms[room_id].state[event_type][state_key],
+            query_params={"format": format} if format != "content" else None,
             metrics_method="getStateEvent",
         )
-        content["__mautrix_event_type"] = event_type
         try:
-            return StateEvent.deserialize_content(content)
+            if format == "content":
+                content["__mautrix_event_type"] = event_type
+                return StateEvent.deserialize_content(content)
+            elif format == "event":
+                return StateEvent.deserialize(content)
+            else:
+                return content
         except SerializerError as e:
             raise MatrixResponseError("Invalid state event in response") from e
 
@@ -263,7 +344,7 @@ class EventMethods(BaseClientAPI):
         self,
         room_id: RoomID,
         direction: PaginationDirection,
-        from_token: SyncToken,
+        from_token: SyncToken | None = None,
         to_token: SyncToken | None = None,
         limit: int | None = None,
         filter_json: str | dict | RoomEventFilter | None = None,
@@ -272,7 +353,7 @@ class EventMethods(BaseClientAPI):
         Get a list of message and state events for a room. Pagination parameters are used to
         paginate history in the room.
 
-        See also: `API reference <https://spec.matrix.org/v1.1/client-server-api/#get_matrixclientv3roomsroomidmessages>`__
+        See also: `API reference <https://spec.matrix.org/v1.3/client-server-api/#get_matrixclientv3roomsroomidmessages>`__
 
         Args:
             room_id: The ID of the room to get events from.
@@ -280,6 +361,9 @@ class EventMethods(BaseClientAPI):
             from_token: The token to start returning events from. This token can be obtained from a
                 ``prev_batch`` token returned for each room by the `sync endpoint`_, or from a
                 ``start`` or ``end`` token returned by a previous request to this endpoint.
+
+                Starting from Matrix v1.3, this field can be omitted to fetch events from the
+                beginning or end of the room.
             to_token: The token to stop returning events at.
             limit: The maximum number of events to return. Defaults to 10.
             filter_json: A JSON RoomEventFilter_ to filter returned events with.
@@ -287,9 +371,9 @@ class EventMethods(BaseClientAPI):
         Returns:
 
         .. _RoomEventFilter:
-            https://spec.matrix.org/v1.1/client-server-api/#filtering
+            https://spec.matrix.org/v1.3/client-server-api/#filtering
         .. _sync endpoint:
-            https://spec.matrix.org/v1.1/client-server-api/#get_matrixclientv3sync
+            https://spec.matrix.org/v1.3/client-server-api/#get_matrixclientv3sync
         """
         if isinstance(filter_json, Serializable):
             filter_json = filter_json.json()
@@ -333,6 +417,7 @@ class EventMethods(BaseClientAPI):
         event_type: EventType,
         content: StateEventContent,
         state_key: str = "",
+        ensure_joined: bool = True,
         **kwargs,
     ) -> EventID:
         """
@@ -346,6 +431,8 @@ class EventMethods(BaseClientAPI):
             event_type: The type of state to send.
             content: The content to send.
             state_key: The key for the state to send. Defaults to empty string.
+            ensure_joined: Used by IntentAPI to determine if it should ensure the user is joined
+                before sending the event.
             **kwargs: Optional parameters to pass to the :meth:`HTTPAPI.request` method. Used by
                 :class:`IntentAPI` to pass the timestamp massaging field to
                 :meth:`AppServiceAPI.request`.
