@@ -1,17 +1,21 @@
-# Copyright (c) 2021 Tulir Asokan
+# Copyright (c) 2022 Tulir Asokan
 #
 # This Source Code Form is subject to the terms of the Mozilla Public
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 from __future__ import annotations
 
-from typing import Any
+from typing import Any, Union
 
 from mautrix.api import Method, Path
 from mautrix.errors import MatrixResponseError
 from mautrix.types import (
+    JSON,
     ClaimKeysResponse,
+    CrossSigningKeys,
+    CrossSigningUsage,
     DeviceID,
+    DeviceKeys,
     EncryptionKeyAlgorithm,
     EventType,
     QueryKeysResponse,
@@ -47,7 +51,7 @@ class CryptoMethods(BaseClientAPI):
             raise ValueError("Event type must be a to-device event type")
         await self.api.request(
             Method.PUT,
-            Path.sendToDevice[event_type][self.api.get_txn_id()],
+            Path.v3.sendToDevice[event_type][self.api.get_txn_id()],
             {
                 "messages": {
                     user_id: {
@@ -82,7 +86,7 @@ class CryptoMethods(BaseClientAPI):
     async def upload_keys(
         self,
         one_time_keys: dict[str, Any] | None = None,
-        device_keys: dict[str, Any] | None = None,
+        device_keys: DeviceKeys | dict[str, Any] | None = None,
     ) -> dict[EncryptionKeyAlgorithm, int]:
         """
         Publishes end-to-end encryption keys for the device.
@@ -102,10 +106,14 @@ class CryptoMethods(BaseClientAPI):
         """
         data = {}
         if device_keys:
+            if isinstance(device_keys, Serializable):
+                device_keys = device_keys.serialize()
             data["device_keys"] = device_keys
         if one_time_keys:
+            if isinstance(one_time_keys, Serializable):
+                one_time_keys = one_time_keys.serialize()
             data["one_time_keys"] = one_time_keys
-        resp = await self.api.request(Method.POST, Path.keys.upload, data)
+        resp = await self.api.request(Method.POST, Path.v3.keys.upload, data)
         try:
             return {
                 EncryptionKeyAlgorithm.deserialize(alg): count
@@ -116,6 +124,43 @@ class CryptoMethods(BaseClientAPI):
         except AttributeError as e:
             raise MatrixResponseError("Invalid `one_time_key_counts` field in response.") from e
 
+    async def upload_cross_signing_keys(
+        self,
+        keys: dict[CrossSigningUsage, CrossSigningKeys],
+        auth: dict[str, JSON] | None = None,
+    ) -> None:
+        await self.api.request(
+            Method.POST,
+            Path.v3.keys.device_signing.upload,
+            {f"{usage}_key": key.serialize() for usage, key in keys.items()}
+            | ({"auth": auth} if auth else {}),
+        )
+
+    async def upload_one_signature(
+        self,
+        user_id: UserID,
+        device_id: DeviceID,
+        keys: Union[DeviceKeys, CrossSigningKeys],
+    ) -> None:
+        await self.api.request(
+            Method.POST, Path.v3.keys.signatures.upload, {user_id: {device_id: keys.serialize()}}
+        )
+        # TODO check failures
+
+    async def upload_many_signatures(
+        self,
+        signatures: dict[UserID, dict[DeviceID, Union[DeviceKeys, CrossSigningKeys]]],
+    ) -> None:
+        await self.api.request(
+            Method.POST,
+            Path.v3.keys.signatures.upload,
+            {
+                user_id: {device_id: keys.serialize() for device_id, keys in devices.items()}
+                for user_id, devices in signatures.items()
+            },
+        )
+        # TODO check failures
+
     async def query_keys(
         self,
         device_keys: list[UserID] | set[UserID] | dict[UserID, list[DeviceID]],
@@ -147,7 +192,7 @@ class CryptoMethods(BaseClientAPI):
         }
         if token:
             data["token"] = token
-        resp = await self.api.request(Method.POST, Path.keys.query, data)
+        resp = await self.api.request(Method.POST, Path.v3.keys.query, data)
         return QueryKeysResponse.deserialize(resp)
 
     async def claim_keys(
@@ -171,7 +216,7 @@ class CryptoMethods(BaseClientAPI):
         """
         resp = await self.api.request(
             Method.POST,
-            Path.keys.claim,
+            Path.v3.keys.claim,
             {
                 "timeout": timeout,
                 "one_time_keys": {
