@@ -1,4 +1,4 @@
-# Copyright (c) 2021 Tulir Asokan
+# Copyright (c) 2022 Tulir Asokan
 #
 # This Source Code Form is subject to the terms of the Mozilla Public
 # License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -10,15 +10,17 @@ import olm
 
 from mautrix.types import (
     DeviceID,
+    DeviceKeys,
     EncryptionAlgorithm,
     EncryptionKeyAlgorithm,
     IdentityKey,
+    KeyID,
     SigningKey,
     UserID,
 )
 
-from . import base
 from .sessions import Session
+from .signature import sign_olm
 
 
 class OlmAccount(olm.Account):
@@ -74,19 +76,18 @@ class OlmAccount(olm.Account):
             session.pickle("roundtrip"), passphrase="roundtrip", creation_time=datetime.now()
         )
 
-    def get_device_keys(self, user_id: UserID, device_id: DeviceID) -> Dict[str, Any]:
-        device_keys = {
-            "user_id": user_id,
-            "device_id": device_id,
-            "algorithms": [EncryptionAlgorithm.OLM_V1.value, EncryptionAlgorithm.MEGOLM_V1.value],
-            "keys": {
-                f"{algorithm}:{device_id}": key for algorithm, key in self.identity_keys.items()
+    def get_device_keys(self, user_id: UserID, device_id: DeviceID) -> DeviceKeys:
+        device_keys = DeviceKeys(
+            user_id=user_id,
+            device_id=device_id,
+            algorithms=[EncryptionAlgorithm.OLM_V1, EncryptionAlgorithm.MEGOLM_V1],
+            keys={
+                KeyID(algorithm=EncryptionKeyAlgorithm(algorithm), key_id=device_id): key
+                for algorithm, key in self.identity_keys.items()
             },
-        }
-        signature = self.sign(base.canonical_json(device_keys))
-        device_keys["signatures"] = {
-            user_id: {f"{EncryptionKeyAlgorithm.ED25519}:{device_id}": signature}
-        }
+            signatures={},
+        )
+        device_keys.signatures[user_id] = {KeyID.ed25519(device_id): sign_olm(device_keys, self)}
         return device_keys
 
     def get_one_time_keys(
@@ -97,12 +98,12 @@ class OlmAccount(olm.Account):
             self.generate_one_time_keys(new_count)
         keys = {}
         for key_id, key in self.one_time_keys.get("curve25519", {}).items():
-            signature = self.sign(base.canonical_json({"key": key}))
-            keys[f"{EncryptionKeyAlgorithm.SIGNED_CURVE25519}:{key_id}"] = {
+            keys[str(KeyID.signed_curve25519(IdentityKey(key_id)))] = {
                 "key": key,
                 "signatures": {
-                    user_id: {f"{EncryptionKeyAlgorithm.ED25519}:{device_id}": signature}
+                    user_id: {
+                        str(KeyID.ed25519(device_id)): sign_olm({"key": key}, self),
+                    }
                 },
             }
-        self.mark_keys_as_published()
         return keys
