@@ -18,8 +18,9 @@ from mautrix.types import (
     UserID,
 )
 
-from .base import BaseOlmMachine, verify_signature_json
+from .base import BaseOlmMachine
 from .sessions import Session
+from .signature import verify_signature_json
 
 ClaimKeysList = Dict[UserID, Dict[DeviceID, DeviceIdentity]]
 
@@ -60,6 +61,7 @@ class OlmEncryptionMachine(BaseOlmMachine):
         self, users: ClaimKeysList, _force_recreate_session: bool = False
     ) -> None:
         request: Dict[UserID, Dict[DeviceID, EncryptionKeyAlgorithm]] = {}
+        expected_devices = set()
         for user_id, devices in users.items():
             request[user_id] = {}
             for device_id, identity in devices.items():
@@ -67,13 +69,18 @@ class OlmEncryptionMachine(BaseOlmMachine):
                     identity.identity_key
                 ):
                     request[user_id][device_id] = EncryptionKeyAlgorithm.SIGNED_CURVE25519
+                    expected_devices.add((user_id, device_id))
             if not request[user_id]:
                 del request[user_id]
         if not request:
             return
+        request_device_count = len(expected_devices)
         keys = await self.client.claim_keys(request)
+        for server, info in (keys.failures or {}).items():
+            self.log.warning(f"Key claim failure for {server}: {info}")
         for user_id, devices in keys.one_time_keys.items():
             for device_id, one_time_keys in devices.items():
+                expected_devices.discard((user_id, device_id))
                 key_id, one_time_key_data = one_time_keys.popitem()
                 one_time_key = one_time_key_data["key"]
                 identity = users[user_id][device_id]
@@ -90,6 +97,19 @@ class OlmEncryptionMachine(BaseOlmMachine):
                         f"Created new Olm session with {user_id}/{device_id} "
                         f"(OTK ID: {key_id})"
                     )
+        if expected_devices:
+            if request_device_count == 1:
+                raise Exception(
+                    "Key claim response didn't contain key "
+                    f"for queried device {expected_devices.pop()}"
+                )
+            else:
+                self.log.warning(
+                    "Key claim response didn't contain keys for %d out of %d expected devices: %s",
+                    len(expected_devices),
+                    request_device_count,
+                    expected_devices,
+                )
 
     async def send_encrypted_to_device(
         self,
