@@ -5,10 +5,12 @@
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 from __future__ import annotations
 
-from typing import Any
+from typing import Any, AsyncContextManager
 from contextlib import asynccontextmanager
+from contextvars import ContextVar
 import asyncio
 import logging
+import os
 import re
 import sqlite3
 
@@ -23,6 +25,9 @@ from .upgrade import UpgradeTable
 POSITIONAL_PARAM_PATTERN = re.compile(r"\$(\d+)")
 
 
+in_transaction = ContextVar("in_transaction", default=False)
+
+
 class TxnConnection(aiosqlite.Connection):
     def __init__(self, path: str, **kwargs) -> None:
         def connector() -> sqlite3.Connection:
@@ -34,7 +39,11 @@ class TxnConnection(aiosqlite.Connection):
 
     @asynccontextmanager
     async def transaction(self) -> None:
+        if in_transaction.get():
+            yield
+            return
         await self.execute("BEGIN TRANSACTION")
+        token = in_transaction.set(True)
         try:
             yield
         except Exception:
@@ -42,6 +51,8 @@ class TxnConnection(aiosqlite.Connection):
             raise
         else:
             await self.commit()
+        finally:
+            in_transaction.reset(token)
 
     def __execute(self, query: str, *args: Any):
         query = POSITIONAL_PARAM_PATTERN.sub(r"?\1", query)
@@ -81,6 +92,7 @@ class TxnConnection(aiosqlite.Connection):
 
 class SQLiteDatabase(Database):
     scheme = Scheme.SQLITE
+    _parent: SQLiteDatabase | None
     _pool: asyncio.Queue[TxnConnection]
     _stopped: bool
     _conns: int
@@ -103,9 +115,8 @@ class SQLiteDatabase(Database):
             owner_name=owner_name,
             ignore_foreign_tables=ignore_foreign_tables,
         )
+        self._parent = None
         self._path = url.path
-        if self._path.startswith("/"):
-            self._path = self._path[1:]
         self._pool = asyncio.Queue(self._db_args.pop("min_size", 1))
         self._db_args.pop("max_size", None)
         self._stopped = False
@@ -116,6 +127,7 @@ class SQLiteDatabase(Database):
     def _add_missing_pragmas(init_commands: list[str]) -> list[str]:
         has_foreign_keys = False
         has_journal_mode = False
+        has_synchronous = False
         has_busy_timeout = False
         for cmd in init_commands:
             if "PRAGMA" not in cmd:
@@ -124,23 +136,39 @@ class SQLiteDatabase(Database):
                 has_foreign_keys = True
             elif "journal_mode" in cmd:
                 has_journal_mode = True
+            elif "synchronous" in cmd:
+                has_synchronous = True
             elif "busy_timeout" in cmd:
                 has_busy_timeout = True
         if not has_foreign_keys:
             init_commands.append("PRAGMA foreign_keys = ON")
         if not has_journal_mode:
             init_commands.append("PRAGMA journal_mode = WAL")
+        if not has_synchronous and "PRAGMA journal_mode = WAL" in init_commands:
+            init_commands.append("PRAGMA synchronous = NORMAL")
         if not has_busy_timeout:
             init_commands.append("PRAGMA busy_timeout = 5000")
         return init_commands
 
+    def override_pool(self, db: Database) -> None:
+        assert isinstance(db, SQLiteDatabase)
+        self._parent = db
+
     async def start(self) -> None:
+        if self._parent:
+            await super().start()
+            return
         if self._conns:
             raise RuntimeError("database pool has already been started")
         elif self._stopped:
             raise RuntimeError("database pool can't be restarted")
         self.log.debug(f"Connecting to {self.url}")
         self.log.debug(f"Database connection init commands: {self._init_commands}")
+        if os.path.exists(self._path):
+            if not os.access(self._path, os.W_OK):
+                self.log.warning("Database file doesn't seem writable")
+        elif not os.access(os.path.dirname(os.path.abspath(self._path)), os.W_OK):
+            self.log.warning("Database file doesn't exist and directory doesn't seem writable")
         for _ in range(self._pool.maxsize):
             conn = await TxnConnection(self._path, **self._db_args)
             if self._init_commands:
@@ -155,14 +183,21 @@ class SQLiteDatabase(Database):
         await super().start()
 
     async def stop(self) -> None:
+        if self._parent:
+            return
         self._stopped = True
         while self._conns > 0:
             conn = await self._pool.get()
             self._conns -= 1
             await conn.close()
 
+    def acquire_direct(self) -> AsyncContextManager[LoggingConnection]:
+        if self._parent:
+            return self._parent.acquire()
+        return self._acquire()
+
     @asynccontextmanager
-    async def acquire(self) -> LoggingConnection:
+    async def _acquire(self) -> LoggingConnection:
         if self._stopped:
             raise RuntimeError("database pool has been stopped")
         conn = await self._pool.get()
