@@ -98,7 +98,7 @@ class Program:
 
         self.log.info(f"Initializing {self.name} {self.version}")
         try:
-            self.prepare()
+            self.loop.run_until_complete(self._async_prepare())
         except Exception:
             self.log.critical("Unexpected error in initialization", exc_info=True)
             sys.exit(1)
@@ -117,9 +117,10 @@ class Program:
         self.prepare_config()
         self.prepare_log()
         self.check_config()
+        self.init_loop()
 
     @property
-    def _default_base_config(self) -> str:
+    def base_config_path(self) -> str:
         return f"pkg://{self.module}/example-config.yaml"
 
     def prepare_arg_parser(self) -> None:
@@ -133,21 +134,13 @@ class Program:
             metavar="<path>",
             help="the path to your config file",
         )
-        self.parser.add_argument(
-            "-b",
-            "--base-config",
-            type=str,
-            default=self._default_base_config,
-            metavar="<path>",
-            help="the path to the example config (for automatic config updates)",
-        )
         self.parser.add_argument(
             "-n", "--no-update", action="store_true", help="Don't save updated config to disk"
         )
 
     def prepare_config(self) -> None:
         """Pre-init lifecycle method. Extend this if you want to customize config loading."""
-        self.config = self.config_class(self.args.config, self.args.base_config)
+        self.config = self.config_class(self.args.config, self.base_config_path)
         self.load_and_update_config()
 
     def load_and_update_config(self) -> None:
@@ -155,13 +148,10 @@ class Program:
         try:
             self.config.update(save=not self.args.no_update)
         except BaseMissingError:
-            if self.args.base_config != self._default_base_config:
-                print(f"Failed to read base config from {self.args.base_config}")
-            else:
-                print(
-                    "Failed to read base config from the default path "
-                    f"({self._default_base_config}). Maybe your installation is corrupted?"
-                )
+            print(
+                "Failed to read base config from the default path "
+                f"({self.base_config_path}). Maybe your installation is corrupted?"
+            )
             sys.exit(12)
 
     def check_config(self) -> None:
@@ -179,14 +169,16 @@ class Program:
         logging.config.dictConfig(copy.deepcopy(self.config["logging"]))
         self.log = cast(TraceLogger, logging.getLogger("mau.init"))
 
+    async def _async_prepare(self) -> None:
+        self.prepare()
+
     def prepare(self) -> None:
         """
         Lifecycle method where the primary program initialization happens.
         Use this to fill startup_actions with async startup tasks.
         """
-        self.prepare_loop()
 
-    def prepare_loop(self) -> None:
+    def init_loop(self) -> None:
         """Init lifecycle method where the asyncio event loop is created."""
         if uvloop is not None:
             uvloop.install()
@@ -199,6 +191,7 @@ class Program:
         try:
             enabled = self.config["metrics.enabled"]
             listen_port = self.config["metrics.listen_port"]
+            hostname = self.config.get("metrics.hostname", "0.0.0.0")
         except KeyError:
             return
         if not enabled:
@@ -208,12 +201,13 @@ class Program:
                 "Metrics are enabled in config, but prometheus_client is not installed"
             )
             return
-        prometheus.start_http_server(listen_port)
+        prometheus.start_http_server(listen_port, addr=hostname)
 
     def _run(self) -> None:
         signal.signal(signal.SIGINT, signal.default_int_handler)
         signal.signal(signal.SIGTERM, signal.default_int_handler)
 
+        self._stop_task = self.loop.create_future()
         exit_code = 0
         try:
             self.log.debug("Running startup actions...")
@@ -224,7 +218,6 @@ class Program:
                 f"Startup actions complete in {round(end_ts - start_ts, 2)} seconds, "
                 "now running forever"
             )
-            self._stop_task = self.loop.create_future()
             exit_code = self.loop.run_until_complete(self._stop_task)
             self.log.debug("manual_stop() called, stopping...")
         except KeyboardInterrupt:
